// VinciaShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Antenna class 
  
#ifndef VINCIA_VinciaAntenna_H
#define VINCIA_VinciaAntenna_H

#include "Basics.h"
#include "Event.h"
#include "PythiaStdlib.h"
#include "VinciaCommon.h"

namespace Vincia {
  
  // Classes contained in this header
  class Antenna;
  class AntennaSet;
  class Antenna;
  class AntennaqqEmit;
  class AntennaqgEmit;
  class AntennaggEmit;
  class AntennaqgSplit;
  class AntennaggSplit;
  class ThreeInt;
  class FourInt;
  class SixInt;
  
  // the nInt classes are needed as indices for STL maps
  // should be a template, but the code is not that long
  class ThreeInt {
  public:
    // Constructor
  ThreeInt(int i1, int i2, int i3) :
    int1(i1), int2(i2), int3(i3) {}
    // comparison
    bool operator<(const ThreeInt &right) const {
      return (
	      int1 < right.int1 ||
	      (int1 == right.int1 && int2 < right.int2) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 < right.int3) 
	      );
    }
    int int1;
    int int2;
    int int3;
  }; // End class ThreeInt
  
  class FourInt {
  public:
    // Constructor
  FourInt(int i1, int i2, int i3, int i4) :
    int1(i1), int2(i2), int3(i3), int4(i4) {}
    // comparison
    bool operator<(const FourInt &right) const {
      return (
	      int1 < right.int1 ||
	      (int1 == right.int1 && int2 < right.int2) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 < right.int3) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 < right.int4)
	      );
    }
    int int1;
    int int2;
    int int3;
    int int4;
  }; // End class FourInt
  
  class SixInt {
  public:
    // Constructor
  SixInt(int i1, int i2, int i3, int i4, int i5, int i6) :
    int1(i1), int2(i2), int3(i3), int4(i4), int5(i5), int6(i6) {}
    // Return operator
    bool operator<(const SixInt &right) const {
      return (
	      int1 < right.int1 ||
	      (int1 == right.int1 && int2 < right.int2) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 < right.int3) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 < right.int4) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 == right.int4 &&
	       int5 < right.int5) ||
	      (int1 == right.int1 && int2 == right.int2 &&
	       int3 == right.int3 && int4 == right.int4 &&
	       int5 == right.int5 && int6 < right.int6)
	      
	      );
    }
    // maybe this should be private, not sure
    int int1;
    int int2;
    int int3;
    int int4;
    int int5;
    int int6;
  }; // End class SixInt
  

  // Base class Antenna which contains all the shared pieces and 
  // a specialisation for each antenna type which overrides some 
  // of the virtual functions
  class Antenna {
    // contains almost all public functions of Antenna
  public:
    //Bare constructor: name only
    Antenna(string Name="antenna") : name(Name), humanName(Name) {
      useFiniteTerms = true; 
      useCollinearTerms = true;
      isInitPtr = false; 
      isOnSav = false;
    };
    
    // Destructor (not sure why the one in Antenna is virtual)
    ~Antenna() {};
    
    // Initialize, does not need to be virtual, just needs to call
    // some virtual functions
    bool init(bool checkAntenna=true);
    void initPtr(ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, 
		 Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    }
    // make a version of initialise with pointers
    bool init(
      bool checkAntenna, ParticleData* particleDataPtrIn, 
      Rndm* rndmPtrIn, Settings* settingsPtrIn
      ) {
      initPtr(particleDataPtrIn, rndmPtrIn, settingsPtrIn);
      return init(checkAntenna);
    }
    bool setC(int i=0,int j=0,double c=0.0); // Set C[i][j]
    
    bool setParm(bool&, string);
    bool setParm(int&, string);
    bool setParm(double&, string);
    bool setParm(string&, string);
    void setHumanName(string humanNameIn) {humanName = humanNameIn;}
    
    // Functions to "get"
    bool isOn() {return (isInit ? isOnSav : false);}
    string getHumanName() { return humanName; }
    string getName() { return name; }
    int getInputMode() {return inputMode;}
    int getEvolutionType() {return evolutionType;};     
    int getKineMapType() {return kineMapType;}  
    double getChargeFactor() {return chargeFactor;}  
    int getCutoffType() { return cutoffType; }                 
    double getCutoffScale() { return cutoffScale; }
    
    double a2ofS(
      double sa1,double s1b, double s, double mass1=0., double mass2=0.,
      double mass3=0., bool forcepositive=true
      ) {
      return (s != 0) ? a2(sa1/s, s1b/s, mass1/sqrt(s), mass2/sqrt(s), 
			   mass3/sqrt(s), forcepositive) : 0.0;
    };
    
    // positivity checks... are common to all antennae
    double a2(
      double ya1,double y1b, double mu1=0., double mu2=0., 
      double mu3=0., bool forcepositive=true);
      
    // The finite part polynomial coefficients 
    // (cannot be made static)
    virtual void readC();              // Read all coefficients
    double getC(int i=0, int j=0);     // Return C[i][j]
    
    virtual void clear();            // Clear all coefficients
    
    // check combines IsPositive and IsFreeOfDeadZones
    bool check(int nPointsCheck, double massmu1, double massmu2, 
	       double massmu3, bool dorepair=false);
    // MR 21.12.2010: IsPositive checks whether an antenna is positive
    // given a set of masses. It has an optional argument dorepair 
    // (default false) which governs whether the finite coefficients 
    // are changed to make the antenna positive.
    bool IsPositive( int nPointsCheck, double mu1, double mu2, 
		     double mu3, bool dorepair = false );
    // MR 21.12.2010: IsFreeOfDeadZones checks whether an antenna has 
    // dead zones given a set of masses. It has an optional argument 
    // dorepair (default false) which governs whether the finite 
    // coefficients are changed to make the dead zones go away.
    bool IsFreeOfDeadZones( 
      int nPointsCheck, double mu1, double mu2, double mu3, 
      bool dorepair = false );
    // check function called by init
    bool checkInit(vector< vector<double> > testMasses);
      
    // MR 16.7.2010: list the coefficients of the antenna
    virtual void listCoefficients(ostream& os=cout);
    
    // rescale finite terms
    virtual bool rescaleFiniteTerms(double rescalingFactor);
               
    // get test masses used in Initialization
    virtual vector< vector<double> > gettestMasses();
    
  protected:
    // core of a2, no checks or similar, third invariant already 
    // calculated by a2
    virtual double a2Raw(
      double ya1, double yab, double y1b, 
      double mu1, double mu2, double mu3);
    
    // Antenna Parameters 
    int evolutionType;
    int kineMapType;
    double chargeFactor;
    
    // Cutoff
    int cutoffType;
    double cutoffScale;

    // Antenna function name 
    string name;       // in xml/cmnd files (e.g., Vincia:qq:Emit)
    string humanName;  // human readable format, derived.
    
    // In- and Out- particles in PDG code format
    vector<int> idOld;
    vector<int> idNew;
    vector<string> nameOld;
    vector<string> nameNew;
    
    // Coefficients of the antenna function 
    // (There is only one possible input mode for now: Laurent)
    int inputMode;  
    // Coefficients of Laurent series
    map<int, map<int, double> > LaurentC;
    
    // Flag whether to use finite terms or not
    bool useFiniteTerms, useCollinearTerms;
    // MR 2.3.2011: Flag whether to use mass terms or not (for testing)
    bool useMassTerms;
    // Dead Zone Avoidance Factor
    double deadZoneAvoidance;
    
    // Initialized flag
    bool isInit;

    // On/off flag
    bool isOnSav;
    
    // Pointers to Pythia-8 classes
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
    
    // Verbose setting
    int verbose;
  };
  
  class AntennaqqEmit : public Antenna {
    public:
      // constructor
      AntennaqqEmit(string nameIn = "Vincia:qQ:emit") : Antenna(nameIn) 
        {};
      // read all coefficients from the settings
      void readC();
      
      // clear all coefficients
      void clear();
      
      void label(ostream& os=cout);
      
      void listCoefficients(ostream& os=cout);
      
      bool rescaleFiniteTerms(double rescalingFactor);
      
      double getM(int mu1ord, int mu3ord, int y12ord, int y23ord);
      bool setM(
        int mu1ord, int mu3ord, int y12ord, int y23ord, double mIn);
      
      double getDenominatorMassPrefactor();
      bool setDenominatorMassPrefactor(double pfIn);
      
      vector< vector<double> > gettestMasses();
    private:
    
      double a2Raw(
        double y12, double y13, double y23, double mu1=0., 
        double mu2=0., double mu3=0.);
    
      double denominatorMassPrefactor;
      map<FourInt, double> LaurentM;
  };
  
  class AntennaqgEmit : public Antenna {
    public:
      // constructor
      AntennaqgEmit(string nameIn = "Vincia:qg:emit") : Antenna(nameIn) 
        {};
      // read all coefficients from the settings
      void readC();
      
      // clear all coefficients
      void clear();
      
      void label(ostream& os=cout);
      
      void listCoefficients(ostream& os=cout);
      
      bool rescaleFiniteTerms(double rescalingFactor);
      
      double getCollinearGluonSplitting();
      bool setCollinearGluonSplitting(double);
      
      double getM(int muord, int y12ord, int y23ord);
      bool setM(int muord, int y12ord, int y23ord, double mIn);
      
      double getMtilde(int mu1ord, int y12ord, int y13ord, int y23ord);
      bool setMtilde(
        int mu1ord, int y12ord, int y13ord, int y23ord, double mIn);
        
      vector< vector<double> > gettestMasses();
    private:
    
      double a2Raw(
        double y12, double y13, double y23, double mu1=0., 
        double mu2=0., double mu3=0.);
        
      map<ThreeInt, double> LaurentM;
      map<FourInt, double> LaurentMtilde;
      double collinearGluonSplitting;
  };
  
  class AntennaggEmit : public Antenna {
    public:
      // constructor
      AntennaggEmit(string nameIn = "Vincia:gg:emit") : Antenna(nameIn) 
        {};
      // read all coefficients from the settings
      void readC();
      
      // clear all coefficients
      void clear();
      
      void label(ostream& os=cout);
      
      void listCoefficients(ostream& os=cout);
      
      double getCollinearGluonSplitting();
      bool setCollinearGluonSplitting(double);
      
    private:
    
      double a2Raw(
        double y12, double y13, double y23, double mu1=0., 
        double mu2=0., double mu3=0.);
        
      double collinearGluonSplitting;
  };
  
  class AntennaqgSplit : public Antenna {
    public:
      // constructor
      AntennaqgSplit(string nameIn = "Vincia:qg:split") : Antenna(nameIn) 
        {};
      // read all coefficients from the settings
      void readC();
      
      // clear all coefficients
      void clear();
      
      void label(ostream& os=cout);
      
      void listCoefficients(ostream& os=cout);
      
      bool rescaleFiniteTerms(double rescalingFactor);
      
      double getM(int muqord, int muqpord, int y12ord, int y23ord);
      bool setM(
        int muqord, int muqpord, int y12ord, int y23ord, double mIn);
        
      double getMtilde(int muqord, int muqpord, int y12ord, int y23ord);
      bool setMtilde(
        int muqord, int muqpord, int y12ord, int y23ord, double mIn);
        
      double getCtilde(int y12ord, int y23ord);
      bool setCtilde(int y12ord, int y23ord, double cIn);
      
      vector< vector<double> > gettestMasses();
    private:
    
      double a2Raw(
        double y12, double y13, double y23, double mu1=0., 
        double mu2=0., double mu3=0.);
      
      map<FourInt, double> LaurentM;
      map<FourInt, double> LaurentMtilde;
      map< pair<int, int>, double > LaurentCtilde;
  };
  
  class AntennaggSplit : public Antenna {
    public:
      // constructor
      AntennaggSplit(string nameIn = "Vincia:gg:split") : Antenna(nameIn) 
        {};
      // read all coefficients from the settings
      void readC();
      
      // clear all coefficients
      void clear();
      
      void label(ostream& os=cout);
      
      void listCoefficients(ostream& os=cout);
      
      bool rescaleFiniteTerms(double rescalingFactor);
      
      double getM(int muqpord, int y12ord, int y23ord);
      bool setM(
        int muqpord, int y12ord, int y23ord, double mIn);
        
      double getCtilde(int y12ord, int y23ord);
      bool setCtilde(int y12ord, int y23ord, double cIn);
      
      vector< vector<double> > gettestMasses();
    private:
      
      double a2Raw(
        double y12, double y13, double y23, double mu1=0., 
        double mu2=0., double mu3=0.);
        
      map<ThreeInt, double> LaurentM;
      map< pair<int, int>, double> LaurentCtilde;
  };
  
  // AntennaSet = List of Antennae 
  class AntennaSet {

  public:
    // destructor
    // delete the antennae
    ~AntennaSet(){
      for (
        vector<Antenna*>::iterator antPtIt = antPtrs.begin(); 
        antPtIt != antPtrs.end(); ++antPtIt
        ) {
        delete *antPtIt;
      }
    }
    // initialisation
    void init(int verbose, bool checkAntennae=false) {
      if (! isInitPtr) {
        printOut("AntennaSet::init",
                  "Cannot initialize, pointers not set.");
        return;
      }
      // Read in and initialize generic Sudakovs
      antPtrs.resize(5);
      antPtrs[iqqEmit] = new AntennaqqEmit();
      antPtrs[iqgEmit] = new AntennaqgEmit();
      antPtrs[iggEmit] = new AntennaggEmit();
      antPtrs[iqgSplit] = new AntennaqgSplit();
      antPtrs[iggSplit] = new AntennaggSplit();
      // loop through antPtrs and initialise
      for (
        vector<Antenna*>::iterator antPtIt = antPtrs.begin(); 
        antPtIt != antPtrs.end(); ++antPtIt
        ) {
        // initialise pointers for antennae, giving pointers as args
        bool pass = 
          (*antPtIt)->init(
            checkAntennae, particleDataPtr, rndmPtr, settingsPtr);
        if (pass) {	
          // Verbose output
          if (verbose >= 3)
            printOut((*antPtIt)->getHumanName()+":init",
                                          "ok: added to antenna list");
        } 
        else if (verbose >= 1) {
            printErr((*antPtIt)->getName()+":init",
              (*antPtIt)->getName()+
                  " has a problem and will not be used in this run!");
            *antPtIt = NULL;
        }
      }
    };
    void initPtr(ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, 
		 Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    }
    
    // at present this does only allow to set generic Antennae,
    // Antennae with a derived type will be converted to Antenna
    void setAnt(vector<Antenna*>::size_type iAntIn, Antenna antIn) {
      if (iAntIn > antPtrs.size()) {
        antPtrs.resize(iAntIn);
      }
      Antenna* antCopyPtr = new Antenna(antIn);
      antPtrs[iAntIn] = antCopyPtr;
      antIndxSav[antCopyPtr->getName()]=iAntIn;
    }
    
    bool exists(int iAntIn) {
      return (iAntIn<0 || iAntIn >= int(antPtrs.size())) ? false : true;
    }
    Antenna* getAnt(vector<Antenna*>::size_type iAntIn) {
      if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn];
      }
      else return NULL;
    }
    string getName(vector<Antenna*>::size_type iAntIn) {
      if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn]->getName();
      }
      else return "none";
    }
    string getHumanName(vector<Antenna*>::size_type iAntIn) {
       if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn]->getHumanName();
      }
      else return "none";
    }
    int getIndx(string antNameIn) {return antIndxSav[antNameIn];}
    
  private:
    // use a vector of Antenna pointers, create them with new on 
    // initialization
    vector<Antenna*> antPtrs;    
    map<string,int> antIndxSav;
    // Pointers to Pythia-8 classes, needed to initialise antennae
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
  }; // End class AntennaSet
  
} // End Vincia namespace

#endif
